###

using LinearAlgebra
using HomotopyContinuation
using DynamicPolynomials
using Dates

@var v[1:14] # Parametrization of the vertices. (v[1],v[2]) is the first vertex, (v[3],v[4]) the second and so on.
@var w[1:14] # Corresponds to the coefficients of the quartic adjoint in the order x^4+w[1]x^3y+w[2]x^2y^2+w[3]xy^3+...

@polyvar x y

function adjCoeff(u)

  # Input: 7 points in the plane, where (u[1],u[2]) is the first point, (u[3],u[4]) is the second and so on.
  # Output: The adjoint curve of the corresponding heptagon, via Warren's triangulation formula from Proposition 4.7.
  adj = 0
  for i in 2:6 # The heptagon is triangulation into 5 triangles defined by u[1],u[2],u[3], u[1],u[3],u[4], and so on.
    M=[u[2*i-1]-u[1] u[2*i]-u[2];u[2*i+1]-u[1] u[2*i+2]-u[2]]
    detT=det(M)/2 ## The volume of the triangle

    prodt=1
    for j in 2:7 # Calculation of the product of degree 1 polynomials.
      if j!=i && j!=i+1 ## go through all j except 1,i and i+1.
        prodt=prodt*(1-u[2*j-1]*x-u[2*j]*y)
      end
    end
    adj=adj+detT*prodt
  end

  L=coefficients(adj)
  return L
end

###

L1=adjCoeff(v)[1] # We normalize the adjoint curve so that the coefficient of x^4 is 1.
vectAdj=[adjCoeff(v)[2]/L1-w[1]]
for i in 2:14
  global vectAdj
  vectAdj=vcat(vectAdj,[adjCoeff(v)[i+1]/L1-w[i]])
end

# Generate the polynomial system from the proof of Theorem 4.2.

F=System(vectAdj; parameters = w)

# Perform monodromy solve

monres=monodromy_solve(F)

# Certify that these are regular, isolated solutions.

certHept = certify(F,solutions(monres),parameters(monres))
