include("EigenvalueSolver.jl")
using LinearAlgebra
using DynamicPolynomials
using HomotopyContinuation

function diff_i(i, exps, pt)
    w = fill(0, 1, size(exps, 2))
    w[i] = 1
    expsnew = exps .- w
    [exps[j, i] * prod([pt[k]^exps[j, k] for k = 1:size(exps, 2)]) for j = 1:size(exps, 1)]
end

function montoexp(Σ)
    # Converts an array of monomials into an array of exponents.
    exparray = exponents.(Σ)
    M = fill(0, length(Σ), length(exparray[1]))
    for i = 1:length(Σ)
        M[i, :] = exparray[i]
    end
    return M
end

function getVdm(exps, pt)
    vec = [prod([pt[k]^exps[j, k] for k = 1:size(exps, 2)]) for j = 1:size(exps, 1)]
    vec = vec / norm(vec)
end

function removePoint(vv, pt)
    j = 0
    for i = 1:length(vv)
        if norm(vv[i] - pt) < 0.0001
            j = i
        end
    end
    if j == 1
        w = vv[2]
        for i = 3:length(vv)
            w = vcat(w, vv[i])
        end
    else
        w = vv[1]
        for i = 2:length(vv)
            if j != i
                w = vcat(w, vv[i])
            end
        end
    end
    if length(w) == 1
        return [w]
    else
        W = reshape(w, 2, length(vv) - 1)
        ww = [W[:, i] for i = 1:length(vv)-1]
        return ww
    end
end

function getAdj_1115(v, s) # v are the vertices of the polypol. s are the six simple nodes of the quintic.
    @polyvar x[1:2]
    mons1 = monomials((1 + sum(x))^1)
    exps1 = montoexp(mons1)
    mons5 = monomials((1 + sum(x))^5)
    exps5 = montoexp(mons5)
    L1 = fill(0.0 + 0.0 * im, 0, 3)
    L2 = fill(0.0 + 0.0 * im, 0, 3)
    L3 = fill(0.0 + 0.0 * im, 0, 3)
    Q4 = fill(0.0 + 0.0 * im, 0, 21)

    L1 = vcat(L1, transpose(getVdm(exps1, v[1])), transpose(getVdm(exps1, v[4])))
    L2 = vcat(L2, transpose(getVdm(exps1, v[1])), transpose(getVdm(exps1, v[2])))
    L3 = vcat(L3, transpose(getVdm(exps1, v[2])), transpose(getVdm(exps1, v[3])))

    Q4 = vcat(
        Q4,
        transpose(getVdm(exps5, s[1])),
        transpose(getVdm(exps5, s[2])),
        transpose(getVdm(exps5, s[3])),
    )
    Q4 = vcat(
        Q4,
        transpose(getVdm(exps5, s[4])),
        transpose(getVdm(exps5, s[5])),
        transpose(getVdm(exps5, s[6])),
    )
    Q4 = vcat(
        Q4,
        transpose(diff_i(1, exps5, s[1])),
        transpose(diff_i(1, exps5, s[2])),
        transpose(diff_i(1, exps5, s[3])),
    )
    Q4 = vcat(
        Q4,
        transpose(diff_i(1, exps5, s[4])),
        transpose(diff_i(1, exps5, s[5])),
        transpose(diff_i(1, exps5, s[6])),
    )
    Q4 = vcat(
        Q4,
        transpose(diff_i(2, exps5, s[1])),
        transpose(diff_i(2, exps5, s[2])),
        transpose(diff_i(2, exps5, s[3])),
    )
    Q4 = vcat(
        Q4,
        transpose(diff_i(2, exps5, s[4])),
        transpose(diff_i(2, exps5, s[5])),
        transpose(diff_i(2, exps5, s[6])),
    )
    Q4 = vcat(Q4, transpose(getVdm(exps5, v[3])), transpose(getVdm(exps5, v[4])))

    f1 = (mons1'*nullspace(L1))[1] # Opposites
    f2 = (mons1'*nullspace(L2))[1]
    f3 = (mons1'*nullspace(L3))[1]
    f4 = (mons5'*nullspace(Q4))[1]
    u13 = EigenvalueSolver.solve_CI_dense([f1; f3], x; DBD = false, complex = true)
    u14 = EigenvalueSolver.solve_CI_dense([f1; f4], x; DBD = false, complex = true)
    u24 = EigenvalueSolver.solve_CI_dense([f2; f4], x; DBD = false, complex = true)
    u34 = EigenvalueSolver.solve_CI_dense([f3; f4], x; DBD = false, complex = true)

    u14 = deleteat!(u14, findall(ℓ -> norm(ℓ - v[4]) < 1e-6, u14))
    u34 = deleteat!(u34, findall(ℓ -> norm(ℓ - v[3]) < 1e-6, u34))

    MA = fill(0.0 + 0.0 * im, 0, 21)
    MA = vcat(MA, transpose(getVdm(exps5, u13[1])))

    for i = 1:6
        # global MA
        MA = vcat(MA, transpose(getVdm(exps5, s[i])))
    end

    for i = 1:5
        # global MA
        MA = vcat(MA, transpose(getVdm(exps5, u24[i])))
    end

    for i = 1:4
        # global MA
        MA = vcat(MA, transpose(getVdm(exps5, u14[i])))
        MA = vcat(MA, transpose(getVdm(exps5, u34[i])))
    end

    # Let MA be the matrix of the 20 points
    cAdj = nullspace(MA)

    c1 = coefficients(f1)
    β1 = c1[2:end] / c1[1]
    c2 = coefficients(f2)
    β2 = c2[2:end] / c2[1]
    c3 = coefficients(f3)
    β3 = c3[2:end] / c3[1]
    c4 = coefficients(f4)
    γ = c4[2:end] / c4[1]

    α = (cAdj/cAdj[1])[2:end]

    return u13, u14, u24, u34, α, β1, β2, β3, γ, f1, f2, f3, f4
end

### The 1115-case. For a fixed adjoint, we have 74 variables and 74 equations.

function getSystem_1115()

    @var v[1:2, 1:4] sn[1:2, 1:6] u13[1:2, 1] u14[1:2, 1:4] u24[1:2, 1:5] u34[1:2, 1:4] # 48 variables (20+4 times two)
    @var α[1:20] β1[1:2] β2[1:2] β3[1:2] γ[1:20] # discounting the adjoint alpha, we have 26 parameters.

    @polyvar x[1:2]

    mons5 = monomials((1 + sum(x))^5)
    mons2 = monomials((1 + sum(x))^2)
    mons1 = monomials((1 + sum(x))^1)

    Adj = dot([1; α], mons5)
    L1 = dot([1; β1], mons1)
    L2 = dot([1; β2], mons1)
    L3 = dot([1; β3], mons1)
    Q4 = dot([1; γ], mons5)

    global eqns = [convert(Expression, subs(Adj, x => u13[:, 1]))]
    for i = 1:4
        global eqns = push!(eqns, convert(Expression, subs(Adj, x => u14[:, i])))
    end
    for i = 1:5
        global eqns = push!(eqns, convert(Expression, subs(Adj, x => u24[:, i])))
    end
    for i = 1:4
        global eqns = push!(eqns, convert(Expression, subs(Adj, x => u34[:, i])))
    end
    for i = 1:6
        global eqns = push!(eqns, convert(Expression, subs(Adj, x => sn[:, i])))
    end
    eqns = push!(eqns, convert(Expression, subs(L1, x => v[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L1, x => v[:, 4])))
    eqns = push!(eqns, convert(Expression, subs(L2, x => v[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L2, x => v[:, 2])))
    eqns = push!(eqns, convert(Expression, subs(L3, x => v[:, 2])))
    eqns = push!(eqns, convert(Expression, subs(L3, x => v[:, 3])))
    eqns = push!(eqns, convert(Expression, subs(L1, x => u13[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L3, x => u13[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L2, x => u24[:, 5])))
    eqns = push!(eqns, convert(Expression, subs(Q4, x => u24[:, 5])))
    for i = 1:4 # Li
        global eqns = push!(eqns, convert(Expression, subs(L1, x => u14[:, i])))
        eqns = push!(eqns, convert(Expression, subs(L2, x => u24[:, i])))
        eqns = push!(eqns, convert(Expression, subs(L3, x => u34[:, i])))
    end
    for i = 1:6 # the quintic
        global eqns = push!(eqns, convert(Expression, subs(Q4, x => sn[:, i])))
        eqns =
            push!(eqns, convert(Expression, subs(differentiate(Q4, x[1]), x => sn[:, i])))
        eqns =
            push!(eqns, convert(Expression, subs(differentiate(Q4, x[2]), x => sn[:, i])))
    end
    eqns = push!(eqns, convert(Expression, subs(Q4, x => v[:, 3])))
    eqns = push!(eqns, convert(Expression, subs(Q4, x => v[:, 4])))
    for i = 1:4 # the quintic
        global eqns = push!(eqns, convert(Expression, subs(Q4, x => u14[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Q4, x => u24[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Q4, x => u34[:, i])))
    end

    ### All 74 equations are done.

    bigsys = System(eqns, parameters = α)

    return bigsys
end
