include("EigenvalueSolver.jl")
using Dates

using LinearAlgebra
using DynamicPolynomials
using HomotopyContinuation

function diff_i(i, exps, pt)
    w = fill(0, 1, size(exps, 2))
    w[i] = 1
    expsnew = exps .- w
    [exps[j, i] * prod([pt[k]^exps[j, k] for k = 1:size(exps, 2)]) for j = 1:size(exps, 1)]
end


function montoexp(Σ)
    # Converts an array of monomials into an array of exponents.
    exparray = exponents.(Σ)
    M = fill(0, length(Σ), length(exparray[1]))
    for i = 1:length(Σ)
        M[i, :] = exparray[i]
    end
    return M
end

function getVdm(exps, pt)
    vec = [prod([pt[k]^exps[j, k] for k = 1:size(exps, 2)]) for j = 1:size(exps, 1)]
    vec = vec / norm(vec)
end

function removePoint(vv, pt)
    j = 0
    for i = 1:length(vv)
        if norm(vv[i] - pt) < 0.0001
            j = i
        end
    end
    if j == 1
        w = vv[2]
        for i = 3:length(vv)
            w = vcat(w, vv[i])
        end
    else
        w = vv[1]
        for i = 2:length(vv)
            if j != i
                w = vcat(w, vv[i])
            end
        end
    end
    if length(w) == 1
        return [w]
    else
        W = reshape(w, 2, length(vv) - 1)
        ww = [W[:, i] for i = 1:length(vv)-1]
        return ww
    end
end

function getAdj_1214(v, s) # v are the vertices of the polypol. The first three points of s are simple nodes of the quartic
    # the last three points of s are intersections between the conic and the quartic.
    @polyvar x[1:2]
    mons1 = monomials((1 + sum(x))^1)
    exps1 = montoexp(mons1)
    mons2 = monomials((1 + sum(x))^2)
    exps2 = montoexp(mons2)
    mons4 = monomials((1 + sum(x))^4)
    exps4 = montoexp(mons4)
    mons5 = monomials((1 + sum(x))^5)
    exps5 = montoexp(mons5)

    L1 = fill(0.0 + 0.0 * im, 0, 3)
    Co2 = fill(0.0 + 0.0 * im, 0, 6)
    L3 = fill(0.0 + 0.0 * im, 0, 3)
    Qua4 = fill(0.0 + 0.0 * im, 0, 15)

    L1 = vcat(L1, transpose(getVdm(exps1, v[1])), transpose(getVdm(exps1, v[4])))
    L3 = vcat(L3, transpose(getVdm(exps1, v[2])), transpose(getVdm(exps1, v[3])))

    Co2 = vcat(Co2, transpose(getVdm(exps2, v[1])), transpose(getVdm(exps2, v[2])))
    Co2 = vcat(
        Co2,
        transpose(getVdm(exps2, s[4])),
        transpose(getVdm(exps2, s[5])),
        transpose(getVdm(exps2, s[6])),
    )
    Qua4 = vcat(
        Qua4,
        transpose(getVdm(exps4, s[1])),
        transpose(diff_i(1, exps4, s[1])),
        transpose(diff_i(2, exps4, s[1])),
    )
    Qua4 = vcat(
        Qua4,
        transpose(getVdm(exps4, s[2])),
        transpose(diff_i(1, exps4, s[2])),
        transpose(diff_i(2, exps4, s[2])),
    )
    Qua4 = vcat(
        Qua4,
        transpose(getVdm(exps4, s[3])),
        transpose(diff_i(1, exps4, s[3])),
        transpose(diff_i(2, exps4, s[3])),
    )
    Qua4 = vcat(Qua4, transpose(getVdm(exps4, v[3])), transpose(getVdm(exps4, v[4])))
    Qua4 = vcat(
        Qua4,
        transpose(getVdm(exps4, s[4])),
        transpose(getVdm(exps4, s[5])),
        transpose(getVdm(exps4, s[6])),
    )

    f1 = (mons1'*nullspace(L1))[1] # Opposites
    f2 = (mons2'*nullspace(Co2))[1]
    f3 = (mons1'*nullspace(L3))[1]
    f4 = (mons4'*nullspace(Qua4))[1]
    u12 = EigenvalueSolver.solve_CI_dense([f1; f2], x; DBD = false, complex = true)
    u13 = EigenvalueSolver.solve_CI_dense([f1; f3], x; DBD = false, complex = true)
    u14 = EigenvalueSolver.solve_CI_dense([f1; f4], x; DBD = false, complex = true)
    u23 = EigenvalueSolver.solve_CI_dense([f2; f3], x; DBD = false, complex = true)
    u24 = EigenvalueSolver.solve_CI_dense([f2; f4], x; DBD = false, complex = true)
    u34 = EigenvalueSolver.solve_CI_dense([f3; f4], x; DBD = false, complex = true)

    u12 = deleteat!(u12, findall(ℓ -> norm(ℓ - v[1]) < 1e-6, u12))
    u23 = deleteat!(u23, findall(ℓ -> norm(ℓ - v[2]) < 1e-6, u23))
    u34 = deleteat!(u34, findall(ℓ -> norm(ℓ - v[3]) < 1e-6, u34))
    u14 = deleteat!(u14, findall(ℓ -> norm(ℓ - v[4]) < 1e-6, u14))

    MA = fill(0.0 + 0.0 * im, 0, 21)
    MA = vcat(MA, transpose(getVdm(exps5, u13[1])))
    MA = vcat(MA, transpose(getVdm(exps5, u12[1])))
    MA = vcat(MA, transpose(getVdm(exps5, u23[1])))

    for i = 1:8
        # global MA
        MA = vcat(MA, transpose(getVdm(exps5, u24[i])))
    end

    for i = 1:3
        # global MA
        MA = vcat(MA, transpose(getVdm(exps5, s[i])))
    end

    for i = 1:3
        # global MA
        MA = vcat(MA, transpose(getVdm(exps5, u34[i])))
        MA = vcat(MA, transpose(getVdm(exps5, u14[i])))
    end

    # Let MA be the matrix of the 20 points
    cAdj = nullspace(MA)
    α = (cAdj/cAdj[1])[2:end]

    c1 = coefficients(f1)
    β1 = c1[2:end] / c1[1]
    c2 = coefficients(f2)
    γ2 = c2[2:end] / c2[1]
    c3 = coefficients(f3)
    β3 = c3[2:end] / c3[1]
    c4 = coefficients(f4)
    δ4 = c4[2:end] / c4[1]

    return u12, u13, u14, u23, u24, u34, α, β1, γ2, β3, δ4, f1, f2, f3, f4
end

function getSystem_1214()
    ### The 1214-case. For a fixed adjoint, we have 70 variables and 70 equations.

    @var v[1:2, 1:4] sn[1:2, 1:3] u12[1:2, 1:1] u13[1:2, 1:1] u14[1:2, 1:3] u23[1:2, 1:1] u24[
        1:2,
        1:8,
    ] u34[1:2, 1:3] # 48 variables (20+4 times two)
    @var α[1:20] β1[1:2] γ2[1:5] β3[1:2] δ4[1:14] # discounting the adjoint alpha, we have 22 parameters.

    @polyvar x[1:2]

    mons1 = monomials((1 + sum(x))^1)
    mons2 = monomials((1 + sum(x))^2)
    mons4 = monomials((1 + sum(x))^4)
    mons5 = monomials((1 + sum(x))^5)

    Adj = dot([1; α], mons5)
    L1 = dot([1; β1], mons1)
    Co2 = dot([1; γ2], mons2)
    L3 = dot([1; β3], mons1)
    Qua4 = dot([1; δ4], mons4)

    ### Eq adj

    global eqns = [convert(Expression, subs(Adj, x => u12[:, 1]))]
    eqns = push!(eqns, convert(Expression, subs(Adj, x => u13[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Adj, x => u23[:, 1])))

    for i = 1:3
        global eqns = push!(eqns, convert(Expression, subs(Adj, x => sn[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Adj, x => u14[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Adj, x => u34[:, i])))
    end
    for i = 1:8
        global eqns = push!(eqns, convert(Expression, subs(Adj, x => u24[:, i])))
    end

    ### Eq others

    eqns = push!(eqns, convert(Expression, subs(L1, x => v[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L1, x => v[:, 4])))
    eqns = push!(eqns, convert(Expression, subs(L3, x => v[:, 2])))
    eqns = push!(eqns, convert(Expression, subs(L3, x => v[:, 3])))

    eqns = push!(eqns, convert(Expression, subs(Co2, x => v[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Co2, x => v[:, 2])))
    eqns = push!(eqns, convert(Expression, subs(Qua4, x => v[:, 3])))
    eqns = push!(eqns, convert(Expression, subs(Qua4, x => v[:, 4])))

    eqns = push!(eqns, convert(Expression, subs(L1, x => u12[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Co2, x => u12[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Co2, x => u23[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L3, x => u23[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L3, x => u13[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L1, x => u13[:, 1])))

    for i = 1:3 # the quintic
        global eqns = push!(eqns, convert(Expression, subs(L3, x => u34[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Qua4, x => u34[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Qua4, x => u14[:, i])))
        eqns = push!(eqns, convert(Expression, subs(L1, x => u14[:, i])))
    end
    for i = 1:8 # the quintic
        global eqns = push!(eqns, convert(Expression, subs(Co2, x => u24[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Qua4, x => u24[:, i])))
    end

    eqns = push!(eqns, convert(Expression, subs(Qua4, x => sn[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Qua4, x => sn[:, 2])))
    eqns = push!(eqns, convert(Expression, subs(Qua4, x => sn[:, 3])))
    eqns = push!(eqns, convert(Expression, subs(differentiate(Qua4, x[1]), x => sn[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(differentiate(Qua4, x[1]), x => sn[:, 2])))
    eqns = push!(eqns, convert(Expression, subs(differentiate(Qua4, x[1]), x => sn[:, 3])))
    eqns = push!(eqns, convert(Expression, subs(differentiate(Qua4, x[2]), x => sn[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(differentiate(Qua4, x[2]), x => sn[:, 2])))
    eqns = push!(eqns, convert(Expression, subs(differentiate(Qua4, x[2]), x => sn[:, 3])))

    ### All 70 equations are done.

    bigsys = System(eqns, parameters = α)

    return bigsys
end
