using LinearAlgebra
using DynamicPolynomials
using HomotopyContinuation
include("EigenvalueSolver.jl")

function diff_i(i, exps, pt)
    w = fill(0, 1, size(exps, 2))
    w[i] = 1
    expsnew = exps .- w
    [exps[j, i] * prod([pt[k]^exps[j, k] for k = 1:size(exps, 2)]) for j = 1:size(exps, 1)]
end


function montoexp(Σ)
    # Converts an array of monomials into an array of exponents.
    exparray = exponents.(Σ)
    M = fill(0, length(Σ), length(exparray[1]))
    for i = 1:length(Σ)
        M[i, :] = exparray[i]
    end
    return M
end

function getVdm(exps, pt)
    vec = [prod([pt[k]^exps[j, k] for k = 1:size(exps, 2)]) for j = 1:size(exps, 1)]
    vec = vec / norm(vec)
end

function removePoint(vv, pt)
    j = 0
    for i = 1:length(vv)
        if norm(vv[i] - pt) < 0.0001
            j = i
        end
    end
    if j == 1
        w = vv[2]
        for i = 3:length(vv)
            w = vcat(w, vv[i])
        end
    else
        w = vv[1]
        for i = 2:length(vv)
            if j != i
                w = vcat(w, vv[i])
            end
        end
    end
    if length(w) == 1
        return [w]
    else
        W = reshape(w, 2, length(vv) - 1)
        ww = [W[:, i] for i = 1:length(vv)-1]
        return ww
    end
end

function getAdj_1232(v, s) # the first s is a simple nodes of the cubic. s[2] common points of the conics.
                           # s[3],s[4] common among first conic and cubic. s[5],s[6] commong among the second conic and cubic.
  @polyvar x[1:2]
  mons1 = monomials((1+sum(x))^1)
  exps1 = montoexp(mons1)
  mons2 = monomials((1+sum(x))^2)
  exps2 = montoexp(mons2)
  mons3 = monomials((1+sum(x))^3)
  exps3 = montoexp(mons3)
  mons5 = monomials((1+sum(x))^5)
  exps5 = montoexp(mons5)
  L1 = fill(0.0 + 0.0 * im, 0, 3)
  Co2 = fill(0.0 + 0.0 * im, 0, 6)
  Cu3 = fill(0.0 + 0.0 * im, 0, 10)
  Co4 = fill(0.0 + 0.0 * im, 0, 6)

  L1 = vcat(L1, transpose(getVdm(exps1,v[1])), transpose(getVdm(exps1,v[4])))

  Co2 = vcat(Co2, transpose(getVdm(exps2,v[1])), transpose(getVdm(exps2,v[2])))
  Co2 = vcat(Co2, transpose(getVdm(exps2,s[2])), transpose(getVdm(exps2,s[3])), transpose(getVdm(exps2,s[4])))
  Cu3 = vcat(Cu3, transpose(getVdm(exps3,s[1])), transpose(diff_i(1, exps3, s[1])), transpose(diff_i(2, exps3, s[1])))
  Cu3 = vcat(Cu3, transpose(getVdm(exps3,v[2])), transpose(getVdm(exps3,v[3])))
  Cu3 = vcat(Cu3, transpose(getVdm(exps3,s[3])), transpose(getVdm(exps3,s[4])))
  Cu3 = vcat(Cu3, transpose(getVdm(exps3,s[5])), transpose(getVdm(exps3,s[6])))
  Co4 = vcat(Co4, transpose(getVdm(exps2,v[3])), transpose(getVdm(exps2,v[4])))
  Co4 = vcat(Co4, transpose(getVdm(exps2,s[2])), transpose(getVdm(exps2,s[5])), transpose(getVdm(exps2,s[6])))

  f1=(mons1'*nullspace(L1))[1] # Opposites
  f2=(mons2'*nullspace(Co2))[1]
  f3=(mons3'*nullspace(Cu3))[1]
  f4=(mons2'*nullspace(Co4))[1]
  u12=EigenvalueSolver.solve_CI_dense([f1;f2],x;DBD = false,complex = true)
  u13=EigenvalueSolver.solve_CI_dense([f1;f3],x;DBD = false,complex = true)
  u14=EigenvalueSolver.solve_CI_dense([f1;f4],x;DBD = false,complex = true)
  u23 = EigenvalueSolver.solve_CI_dense([f2;f3],x;DBD = false,complex = true)
  u24 = EigenvalueSolver.solve_CI_dense([f2;f4],x;DBD = false,complex = true)
  u34 = EigenvalueSolver.solve_CI_dense([f3;f4],x;DBD = false,complex = true)

  u12=deleteat!(u12,findall(ℓ-> norm(ℓ - v[1]) < 1e-6, u12))
  u23=deleteat!(u23,findall(ℓ-> norm(ℓ - v[2]) < 1e-6, u23))
  u34=deleteat!(u34,findall(ℓ-> norm(ℓ - v[3]) < 1e-6, u34))
  u14=deleteat!(u14,findall(ℓ-> norm(ℓ - v[4]) < 1e-6, u14))

  MA = fill(0.0 + 0.0 * im, 0, 21)
  MA = vcat(MA, transpose(getVdm(exps5,s[1])))
  MA = vcat(MA, transpose(getVdm(exps5,u14[1])))
  MA = vcat(MA, transpose(getVdm(exps5,u12[1])))

  for i in 1:5
    # global MA
    MA = vcat(MA, transpose(getVdm(exps5,u23[i])))
    MA = vcat(MA, transpose(getVdm(exps5,u34[i])))
  end

  for i in 1:3
    # global MA
    MA = vcat(MA, transpose(getVdm(exps5,u13[i])))
  end

  for i in 1:4
    # global MA
    MA = vcat(MA, transpose(getVdm(exps5,u24[i])))
  end

  # Let MA be the matrix of the 20 points
  cAdj = nullspace(MA)
  α = (cAdj/cAdj[1])[2:end]

  c1 = coefficients(f1)
  β1 = c1[2:end]/c1[1]
  c2 = coefficients(f2)
  γ2 = c2[2:end]/c2[1]
  c3 = coefficients(f3)
  δ3 = c3[2:end]/c3[1]
  c4 = coefficients(f4)
  γ4 = c4[2:end]/c4[1]

  return u12, u13, u14, u23, u24, u34, α, β1, γ2, δ3, γ4, f1, f2, f3, f4
end

function getSystem_1232()
    ### The 1232-case. For a fixed adjoint, we have 69 variables and 69 equations.

    @var v[1:2, 1:4] sn[1:2, 1:1] u12[1:2, 1:1] u13[1:2, 1:3] u14[1:2, 1:1] u23[1:2, 1:5] u24[
        1:2,
        1:4,
    ] u34[1:2, 1:5] # 48 variables (20+4 times two)
    @var α[1:20] β1[1:2] γ2[1:5] δ3[1:9] γ4[1:5] # discounting the adjoint alpha, we have 22 parameters.

    # Note u12 empty.

    @polyvar x[1:2]

    mons5 = monomials((1 + sum(x))^5)
    mons3 = monomials((1 + sum(x))^3)
    mons2 = monomials((1 + sum(x))^2)
    mons1 = monomials((1 + sum(x))^1)

    Adj = dot([1; α], mons5)
    L1 = dot([1; β1], mons1)
    Co2 = dot([1; γ2], mons2)
    Cu3 = dot([1; δ3], mons3)
    Co4 = dot([1; γ4], mons2)

    ### Eq adj

    eqns = [convert(Expression, subs(Adj, x => sn[:, 1]))]
    eqns = push!(eqns, convert(Expression, subs(Adj, x => u12[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Adj, x => u14[:, 1])))

    for i = 1:3
        eqns = push!(eqns, convert(Expression, subs(Adj, x => u13[:, i])))
    end
    for i = 1:4
        eqns = push!(eqns, convert(Expression, subs(Adj, x => u24[:, i])))
    end
    for i = 1:5
        eqns = push!(eqns, convert(Expression, subs(Adj, x => u23[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Adj, x => u34[:, i])))
    end

    ### Eq others

    eqns = push!(eqns, convert(Expression, subs(L1, x => v[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L1, x => v[:, 4])))
    eqns = push!(eqns, convert(Expression, subs(Co2, x => v[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Co2, x => v[:, 2])))

    eqns = push!(eqns, convert(Expression, subs(Cu3, x => v[:, 2])))
    eqns = push!(eqns, convert(Expression, subs(Cu3, x => v[:, 3])))
    eqns = push!(eqns, convert(Expression, subs(Co4, x => v[:, 3])))
    eqns = push!(eqns, convert(Expression, subs(Co4, x => v[:, 4])))

    eqns = push!(eqns, convert(Expression, subs(L1, x => u12[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Co2, x => u12[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(L1, x => u14[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(Co4, x => u14[:, 1])))

    for i = 1:3 # the quintic
        eqns = push!(eqns, convert(Expression, subs(L1, x => u13[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Cu3, x => u13[:, i])))
    end
    for i = 1:4 # the quintic
        eqns = push!(eqns, convert(Expression, subs(Co2, x => u24[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Co4, x => u24[:, i])))
    end
    for i = 1:5 # the quintic
        eqns = push!(eqns, convert(Expression, subs(Co2, x => u23[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Cu3, x => u23[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Cu3, x => u34[:, i])))
        eqns = push!(eqns, convert(Expression, subs(Co4, x => u34[:, i])))
    end

    eqns = push!(eqns, convert(Expression, subs(Cu3, x => sn[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(differentiate(Cu3, x[1]), x => sn[:, 1])))
    eqns = push!(eqns, convert(Expression, subs(differentiate(Cu3, x[2]), x => sn[:, 1])))

    ### All 69 equations are done.

    bigsys = System(eqns, parameters = α)
    return bigsys
end
