using LinearAlgebra
using DynamicPolynomials
using HomotopyContinuation

function montoexp(Σ)
    # Converts an array of monomials into an array of exponents.
    exparray = exponents.(Σ)
    M = fill(0, length(Σ), length(exparray[1]))
    for i = 1:length(Σ)
        M[i, :] = exparray[i]
    end
    return M
end

function getVdm(exps, pt)
    [prod([pt[k]^exps[j, k] for k = 1:size(exps, 2)]) for j = 1:size(exps, 1)]
end

function removePoint(vv, pt)
  j=0
  for i in 1:length(vv)
    if norm(vv[i]-pt)<0.0001
      j=i
    end
  end
  if j==1
    w=vv[2]
    for i in 3:length(vv)
      w=vcat(w,vv[i])
    end
  else
    w=vv[1]
    for i in 2:length(vv)
      if j!=i
      w=vcat(w,vv[i])
      end
    end
  end
  if length(w)==1
    return [w]
  else
    W = reshape(w, 2, length(vv)-1)
    ww = [W[:, i] for i = 1:length(vv)-1]
    return ww
  end
end

function getAdj_2222(v, a, b)
  @polyvar x[1:2]
  mons2 = DynamicPolynomials.monomials((1+sum(x))^2)
  exps2 = montoexp(mons2)
  mons5 = DynamicPolynomials.monomials((1+sum(x))^5)
  exps5 = montoexp(mons5)
  M1 = fill(0.0 + 0.0 * im, 0, 6)
  M2 = fill(0.0 + 0.0 * im, 0, 6)
  M3 = fill(0.0 + 0.0 * im, 0, 6)
  M4 = fill(0.0 + 0.0 * im, 0, 6)

  M1 = vcat(M1, transpose(getVdm(exps2,v[1])), transpose(getVdm(exps2,v[2])))
  M1 = vcat(M1, transpose(getVdm(exps2,a[1])), transpose(getVdm(exps2,a[2])), transpose(getVdm(exps2,a[3])))

  M2 = vcat(M2, transpose(getVdm(exps2,v[2])), transpose(getVdm(exps2,v[4])))
  M2 = vcat(M2, transpose(getVdm(exps2,a[1])), transpose(getVdm(exps2,a[2])), transpose(getVdm(exps2,a[3])))

  M3 = vcat(M3, transpose(getVdm(exps2,v[4])), transpose(getVdm(exps2,v[3])))
  M3 = vcat(M3, transpose(getVdm(exps2,b[1])), transpose(getVdm(exps2,b[2])), transpose(getVdm(exps2,b[3])))

  M4 = vcat(M4, transpose(getVdm(exps2,v[1])), transpose(getVdm(exps2,v[3])))
  M4 = vcat(M4, transpose(getVdm(exps2,b[1])), transpose(getVdm(exps2,b[2])), transpose(getVdm(exps2,b[3])))

  f1=dot(mons2,nullspace(M1)) # Opposites
  f3=dot(mons2,nullspace(M3))
  result13=solve(System([f1,f3]))
  z = solutions(result13)

  f2=dot(mons2,nullspace(M2)) # Opposites
  f4=dot(mons2,nullspace(M4))
  result24=solve(System([f2,f4]))
  y = solutions(result24)

  result14=solve(System([f1,f4])) # From here we need to remove the two vertices
  w = solutions(result14)
  result23=solve(System([f2,f3]))
  u = solutions(result23)

  w=removePoint(w,v[1])
  u=removePoint(u,v[4])

  MA = fill(0.0 + 0.0 * im, 0, 21)
  MA = vcat(MA, transpose(getVdm(exps5,a[1])), transpose(getVdm(exps5,a[2])), transpose(getVdm(exps5,a[3])))
  MA = vcat(MA, transpose(getVdm(exps5,b[1])), transpose(getVdm(exps5,b[2])), transpose(getVdm(exps5,b[3])))

  for i in 1:4
    # global MA
    MA = vcat(MA, transpose(getVdm(exps5,z[i])))
    MA = vcat(MA, transpose(getVdm(exps5,y[i])))
  end

  for i in 1:3
    # global MA
    MA = vcat(MA, transpose(getVdm(exps5,w[i])))
    MA = vcat(MA, transpose(getVdm(exps5,u[i])))
  end

  # Let MA be the matrix of the 20 points
  cAdj = nullspace(MA)

  c1 = DynamicPolynomials.coefficients(f1)
  β = c1[2:end]/c1[1]
  c2 = DynamicPolynomials.coefficients(f2)
  γ = c2[2:end]/c2[1]
  c3 = DynamicPolynomials.coefficients(f3)
  δ = c3[2:end]/c3[1]
  c4 = DynamicPolynomials.coefficients(f4)
  ε = c4[2:end]/c4[1]

  α = (cAdj/cAdj[1])[2:end]

  return u, w, z, y, α, β, γ, δ, ε, f1, f2, f3, f4, x
end

function getSystem_2222()
  @var a[1:2,1:3] u[1:2,1:3] b[1:2,1:3] w[1:2,1:3] v[1:2,1:4] z[1:2,1:4] y[1:2,1:4]
  @var α[1:20] β[1:5] γ[1:5] δ[1:5] ϵ[1:5]
  @polyvar x[1:2]

  mons5 = DynamicPolynomials.monomials((1+sum(x))^5)
  mons2 = DynamicPolynomials.monomials((1+sum(x))^2)

  Adj = dot([1;α], mons5)
  C1 = dot([1;β], mons2)
  C2 = dot([1;γ], mons2)
  C3 = dot([1;δ], mons2)
  C4 = dot([1;ϵ], mons2)

  eqns = [convert(Expression,subs(Adj,x=>z[:,1]))]
  for i = 2:4
      eqns = push!(eqns,convert(Expression,subs(Adj,x=>z[:,i])))
  end
  for i = 1:4
      eqns = push!(eqns,convert(Expression,subs(Adj,x=>y[:,i])))
  end
  for i = 1:3
      eqns = push!(eqns,convert(Expression,subs(Adj,x=>u[:,i])))
      eqns = push!(eqns,convert(Expression,subs(Adj,x=>w[:,i])))
      eqns = push!(eqns,convert(Expression,subs(Adj,x=>b[:,i])))
      eqns = push!(eqns,convert(Expression,subs(Adj,x=>a[:,i])))
  end
  for i = 1:3
      eqns = push!(eqns,convert(Expression,subs(C1,x=>w[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C1,x=>a[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C2,x=>a[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C2,x=>u[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C3,x=>b[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C3,x=>u[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C4,x=>b[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C4,x=>w[:,i])))
  end
  for i = 1:4
      eqns = push!(eqns,convert(Expression,subs(C1,x=>z[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C3,x=>z[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C2,x=>y[:,i])))
      eqns = push!(eqns,convert(Expression,subs(C4,x=>y[:,i])))
  end
  eqns = push!(eqns,convert(Expression,subs(C1,x=>v[:,1])))
  eqns = push!(eqns,convert(Expression,subs(C1,x=>v[:,2])))
  eqns = push!(eqns,convert(Expression,subs(C2,x=>v[:,2])))
  eqns = push!(eqns,convert(Expression,subs(C2,x=>v[:,4])))
  eqns = push!(eqns,convert(Expression,subs(C3,x=>v[:,3])))
  eqns = push!(eqns,convert(Expression,subs(C3,x=>v[:,4])))
  eqns = push!(eqns,convert(Expression,subs(C4,x=>v[:,1])))
  eqns = push!(eqns,convert(Expression,subs(C4,x=>v[:,3])))

  bigsys = System(eqns, parameters = α)
  return bigsys
end
